# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Definitions of basic Eve concepts."""


from __future__ import annotations

import ast
import copy
import re

from . import datamodels, exceptions
from . import extended_typing as xtyping
from . import trees, utils
from .datamodels import validators as _validators
from .extended_typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Final,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from .type_definitions import ConstrainedStr, IntEnum, StrEnum


_SYMBOL_NAME_RE: Final = re.compile(r"^[a-zA-Z_]\w*$")


class SymbolName(ConstrainedStr, regex=_SYMBOL_NAME_RE):
    """String value containing a valid symbol name for typical programming conventions."""

    __slots__ = ()


class SymbolRef(ConstrainedStr, regex=_SYMBOL_NAME_RE):
    """Reference to a symbol name."""

    __slots__ = ()


@datamodels.datamodel(slots=True, frozen=True)
class SourceLocation:
    """Source code location (line, column, source)."""

    line: int = datamodels.field(validator=_validators.ge(1))
    column: int = datamodels.field(validator=_validators.ge(1))
    source: str
    end_line: Optional[int] = datamodels.field(validator=_validators.optional(_validators.ge(1)))
    end_column: Optional[int] = datamodels.field(validator=_validators.optional(_validators.ge(1)))

    @classmethod
    def from_AST(cls, ast_node: ast.AST, source: Optional[str] = None) -> SourceLocation:
        if (
            not isinstance(ast_node, ast.AST)
            or getattr(ast_node, "lineno", None) is None
            or getattr(ast_node, "col_offset", None) is None
        ):
            raise ValueError(
                f"Passed AST node '{ast_node}' does not contain a valid source location."
            )
        if source is None:
            source = f"<ast.{type(ast_node).__name__} at 0x{id(ast_node):x}>"
        return cls(
            ast_node.lineno,
            ast_node.col_offset + 1,
            source,
            end_line=ast_node.end_lineno,
            end_column=ast_node.end_col_offset + 1 if ast_node.end_col_offset is not None else None,
        )

    def __init__(
        self,
        line: int,
        column: int,
        source: str,
        *,
        end_line: Optional[int] = None,
        end_column: Optional[int] = None,
    ) -> None:
        assert end_column is None or end_line is not None
        self.__auto_init__(  # type: ignore[attr-defined]  # __auto_init__ added dynamically
            line=line, column=column, source=source, end_line=end_line, end_column=end_column
        )

    def __str__(self) -> str:
        src = self.source or ""

        end_part = ""
        if self.end_line is not None:
            end_part += f" to {self.end_line}"
        if self.end_column is not None:
            end_part += f":{self.end_column}"

        return f"<{src}:{self.line}:{self.column}{end_part}>"


@datamodels.datamodel(slots=True, frozen=True)
class SourceLocationGroup:
    """A group of merged source code locations (with optional info)."""

    locations: Tuple[SourceLocation, ...] = datamodels.field(validator=_validators.non_empty())
    context: Optional[Union[str, Tuple[str, ...]]]

    def __init__(
        self, *locations: SourceLocation, context: Optional[Union[str, Tuple[str, ...]]] = None
    ) -> None:
        self.__auto_init__(locations=locations, context=context)  # type: ignore[attr-defined]  # __auto_init__ added dynamically

    def __str__(self) -> str:
        locs = ", ".join(str(loc) for loc in self.locations)
        context = f"#{self.context}#" if self.context else ""
        return f"<{context}[{locs}]>"


AnySourceLocation = Union[SourceLocation, SourceLocationGroup]

_T = TypeVar("_T")


class AnnexManager:
    register: ClassVar[Dict[str, Any]] = {}

    @classmethod
    def register_user(
        cls: Type[AnnexManager],
        key: str,
        type_hint: xtyping.TypeAnnotation,
        *,
        shared: bool = False,
    ) -> Callable[[_T], _T]:
        assert isinstance(key, str)

        def _decorator(owner: _T) -> _T:
            if key in cls.register:
                reg_shared, reg_type, reg_owner = cls.register[key]
                if not shared:
                    raise exceptions.EveRuntimeError(
                        f"Annex key '{key}' has been already registered by {reg_owner}."
                    )
                if not reg_shared:
                    raise exceptions.EveRuntimeError(
                        f"Annex key '{key}' has been privately registered by {reg_owner}."
                    )
                elif type_hint != reg_type:
                    raise exceptions.EveRuntimeError(
                        f"Annex key '{key}' type '{type_hint}' does not match registered type '{reg_type}' "
                        f"registered by {reg_owner}."
                    )
                owners = reg_owner
            else:
                owners = []

            cls.register[key] = (shared, type_hint, [*owners, owner])

            return owner

        return _decorator


register_annex_user = AnnexManager.register_user


class Node(datamodels.DataModel, trees.Tree, kw_only=True):  # type: ignore[call-arg]  # kw_only from DataModel
    """Base class representing a node in a syntax tree.

    Implemented as a :class:`eve.datamodels.DataModel` with some extra features.

    Field values should be either:

        * builtin types: `bool`, `bytes`, `int`, `float`, `str`
        * enum.Enum types
        * other :class:`Node` subclasses
        * other :class:`eve.datamodels.DataModel` subclasses
        * supported collections (:class:`List`, :class:`Dict`, :class:`Set`)
            of any of the previous items

    The `annex` attribute is used to dynamically add data to a node, even to
    frozen classes. Data in the `annex` do not affect the hash or equality
    comparisons of the node, since it is not really a field. Thus, visitors
    and pipeline passes can freely attach computed attributes into the instance
    `annex`.
    """

    __slots__ = ()

    @property
    def annex(self) -> utils.Namespace:
        if not hasattr(self, "__node_annex__"):
            object.__setattr__(self, "__node_annex__", utils.Namespace())
        return self.__node_annex__  # type: ignore[attr-defined]  # __node_annex__ added dynamically

    def iter_children_values(self) -> Iterable:
        for name in self.__datamodel_fields__.keys():
            yield getattr(self, name)

    def iter_children_items(self) -> Iterable[Tuple[trees.TreeKey, Any]]:
        for name in self.__datamodel_fields__.keys():
            yield name, getattr(self, name)

    pre_walk_items = trees.pre_walk_items
    pre_walk_values = trees.pre_walk_values

    post_walk_items = trees.post_walk_items
    post_walk_values = trees.post_walk_values

    bfs_walk_items = trees.bfs_walk_items
    bfs_walk_values = trees.bfs_walk_values

    walk_items = trees.walk_items
    walk_values = trees.walk_values

    def copy(self: _T, update: Dict[str, Any]) -> _T:
        new_node = copy.deepcopy(self)
        for k, v in update.items():
            setattr(new_node, k, v)
        return new_node

    # TODO(egparedes): add useful hashes to base node
    # # @property
    # def content_id(self) -> int:
    #     ...
    #
    # @property
    # def annex_content_id(self) -> int:
    #     ...
    #
    # @property
    # def node_content_id(self) -> int:
    #     ...
    #
    # @property
    # def instance_content_id(self) -> int:
    #     ...
    #
    # @property
    # def instance_id(self) -> int:
    #     ...


NodeT = TypeVar("NodeT", bound="Node")
ValueNode = Union[bool, bytes, int, float, str, IntEnum, StrEnum]
LeafNode = Union[NodeT, ValueNode]
CollectionNode = Union[List[LeafNode], Dict[Any, LeafNode], Set[LeafNode]]
RootNode = Union[NodeT, CollectionNode]


class FrozenNode(Node, frozen=True):  # type: ignore[call-arg]  # frozen from DataModel
    ...


class GenericNode(datamodels.GenericDataModel, Node, kw_only=True):  # type: ignore[call-arg]  # kw_only from DataModel
    pass


class VType(datamodels.FrozenModel):

    # Unique name
    name: str


def eq_nonlocated(a: Node, b: Node) -> bool:
    """Compare two nodes, ignoring their `SourceLocation` or `SourceLocationGroup`."""
    return len(utils.ddiff(a, b, exclude_types=[SourceLocation, SourceLocationGroup])) == 0



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import re
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import dace
import dace.data
import numpy as np

from gt4py import eve
from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc import daceir as dcir
from gt4py.cartesian.gtc import oir
from gt4py.cartesian.gtc.common import CartesianOffset
from gt4py.cartesian.gtc.passes.oir_optimizations.utils import compute_horizontal_block_extents


def array_dimensions(array: dace.data.Array):
    dims = [
        any(
            re.match(f"__.*_{k}_stride", str(sym))
            for st in array.strides
            for sym in dace.symbolic.pystr_to_symbolic(st).free_symbols
        )
        or any(
            re.match(f"__{k}", str(sym))
            for sh in array.shape
            for sym in dace.symbolic.pystr_to_symbolic(sh).free_symbols
        )
        for k in "IJK"
    ]
    return dims


def replace_strides(arrays, get_layout_map):
    symbol_mapping = {}
    for array in arrays:
        dims = array_dimensions(array)
        ndata_dims = len(array.shape) - sum(dims)
        axes = [ax for ax, m in zip("IJK", dims) if m] + [str(i) for i in range(ndata_dims)]
        layout = get_layout_map(axes)
        if array.transient:
            stride = 1
            for idx in reversed(np.argsort(layout)):
                symbol = array.strides[idx]
                if symbol.is_symbol:
                    symbol_mapping[str(symbol)] = dace.symbolic.pystr_to_symbolic(stride)
                stride *= array.shape[idx]
    return symbol_mapping


def get_tasklet_symbol(name, offset, is_target):
    if is_target:
        return f"__{name}"

    acc_name = name + "__"
    if offset is not None:
        offset_strs = []
        for axis in dcir.Axis.dims_3d():
            off = offset.to_dict()[axis.lower()]
            if off is not None and off != 0:
                offset_strs.append(axis.lower() + ("m" if off < 0 else "p") + f"{abs(off):d}")
        suffix = "_".join(offset_strs)
        if suffix != "":
            acc_name += suffix
    return acc_name


def axes_list_from_flags(flags):
    return [ax for f, ax in zip(flags, dcir.Axis.dims_3d()) if f]


class AccessInfoCollector(eve.NodeVisitor):
    def __init__(self, collect_read: bool, collect_write: bool, include_full_domain: bool = False):
        self.collect_read: bool = collect_read
        self.collect_write: bool = collect_write
        self.include_full_domain: bool = include_full_domain

    @dataclass
    class Context:
        axes: Dict[str, List["dcir.Axis"]]
        access_infos: Dict[str, "dcir.FieldAccessInfo"] = field(default_factory=dict)

    def visit_VerticalLoop(
        self, node: oir.VerticalLoop, *, block_extents, ctx, **kwargs: Any
    ) -> Dict[str, "dcir.FieldAccessInfo"]:
        for section in reversed(node.sections):
            self.visit(section, block_extents=block_extents, ctx=ctx, **kwargs)
        return ctx.access_infos

    def visit_VerticalLoopSection(
        self,
        node: oir.VerticalLoopSection,
        *,
        block_extents,
        ctx,
        grid_subset=None,
        **kwargs: Any,
    ) -> Dict[str, "dcir.FieldAccessInfo"]:
        inner_ctx = self.Context(axes=ctx.axes)

        if grid_subset is None:
            grid_subset = dcir.GridSubset.from_interval(node.interval, dcir.Axis.K)
        elif dcir.Axis.K not in grid_subset.intervals:
            intervals = dict(dcir.GridSubset.from_interval(node.interval, dcir.Axis.K).intervals)
            intervals.update(grid_subset.intervals)
            grid_subset = dcir.GridSubset(intervals=intervals)
        self.visit(
            node.horizontal_executions,
            block_extents=block_extents,
            ctx=inner_ctx,
            grid_subset=grid_subset,
            k_interval=node.interval,
            **kwargs,
        )
        inner_infos = inner_ctx.access_infos

        k_grid = dcir.GridSubset.from_interval(grid_subset.intervals[dcir.Axis.K], dcir.Axis.K)
        inner_infos = {name: info.apply_iteration(k_grid) for name, info in inner_infos.items()}

        ctx.access_infos.update(
            {
                name: info.union(ctx.access_infos.get(name, info))
                for name, info in inner_infos.items()
            }
        )

        return ctx.access_infos

    def visit_HorizontalExecution(
        self,
        node: oir.HorizontalExecution,
        *,
        block_extents,
        ctx: Context,
        k_interval,
        grid_subset=None,
        **kwargs,
    ) -> Dict[str, "dcir.FieldAccessInfo"]:
        horizontal_extent = block_extents(node)

        inner_ctx = self.Context(axes=ctx.axes)
        inner_infos = inner_ctx.access_infos
        ij_grid = dcir.GridSubset.from_gt4py_extent(horizontal_extent)
        he_grid = ij_grid.set_interval(dcir.Axis.K, k_interval)
        self.visit(
            node.body,
            horizontal_extent=horizontal_extent,
            ctx=inner_ctx,
            he_grid=he_grid,
            grid_subset=grid_subset,
            **kwargs,
        )

        if grid_subset is not None:
            for axis in ij_grid.axes():
                if axis in grid_subset.intervals:
                    ij_grid = ij_grid.set_interval(axis, grid_subset.intervals[axis])

        inner_infos = {name: info.apply_iteration(ij_grid) for name, info in inner_infos.items()}

        ctx.access_infos.update(
            {
                name: info.union(ctx.access_infos.get(name, info))
                for name, info in inner_infos.items()
            }
        )

        return ctx.access_infos

    def visit_AssignStmt(self, node: oir.AssignStmt, **kwargs):
        self.visit(node.right, is_write=False, **kwargs)
        self.visit(node.left, is_write=True, **kwargs)

    def visit_HorizontalRestriction(
        self, node: oir.HorizontalRestriction, *, is_conditional=False, **kwargs
    ):
        self.visit(node.mask, is_conditional=is_conditional, **kwargs)
        self.visit(node.body, is_conditional=True, region=node.mask, **kwargs)

    def visit_MaskStmt(self, node: oir.MaskStmt, *, is_conditional=False, **kwargs):

        self.visit(node.mask, is_conditional=is_conditional, **kwargs)
        self.visit(node.body, is_conditional=True, **kwargs)

    def visit_While(self, node: oir.While, *, is_conditional=False, **kwargs):
        self.generic_visit(node, is_conditional=True, **kwargs)

    @staticmethod
    def _global_grid_subset(
        region: common.HorizontalMask, he_grid: "dcir.GridSubset", offset: List[Optional[int]]
    ):
        res: Dict[
            dcir.Axis, Union[dcir.DomainInterval, dcir.TileInterval, dcir.IndexWithExtent]
        ] = {}
        if region is not None:
            for axis, oir_interval in zip(dcir.Axis.dims_horizontal(), region.intervals):
                he_grid_interval = he_grid.intervals[axis]
                assert isinstance(he_grid_interval, dcir.DomainInterval)
                start = (
                    oir_interval.start if oir_interval.start is not None else he_grid_interval.start
                )
                end = oir_interval.end if oir_interval.end is not None else he_grid_interval.end
                dcir_interval = dcir.DomainInterval(
                    start=dcir.AxisBound.from_common(axis, start),
                    end=dcir.AxisBound.from_common(axis, end),
                )
                res[axis] = dcir.DomainInterval.union(dcir_interval, res.get(axis, dcir_interval))
        if dcir.Axis.K in he_grid.intervals:
            off = offset[dcir.Axis.K.to_idx()] or 0
            he_grid_k_interval = he_grid.intervals[dcir.Axis.K]
            assert not isinstance(he_grid_k_interval, dcir.TileInterval)
            res[dcir.Axis.K] = he_grid_k_interval.shifted(off)
        for axis in dcir.Axis.dims_horizontal():
            iteration_interval = he_grid.intervals[axis]
            mask_interval = res.get(axis, iteration_interval)
            res[axis] = dcir.DomainInterval.intersection(
                axis, iteration_interval, mask_interval
            ).shifted(offset[axis.to_idx()])
        return dcir.GridSubset(intervals=res)

    def _make_access_info(
        self,
        offset_node: Union[CartesianOffset, oir.VariableKOffset],
        axes,
        is_conditional,
        region,
        he_grid,
        grid_subset,
    ) -> "dcir.FieldAccessInfo":
        offset = [offset_node.to_dict()[k] for k in "ijk"]
        if isinstance(offset_node, oir.VariableKOffset):
            variable_offset_axes = [dcir.Axis.K]
        else:
            variable_offset_axes = []

        global_subset = self._global_grid_subset(region, he_grid, offset)
        intervals = {}
        for axis in axes:
            if axis in variable_offset_axes:
                intervals[axis] = dcir.IndexWithExtent(
                    axis=axis, value=axis.iteration_symbol(), extent=(0, 0)
                )
            else:
                intervals[axis] = dcir.IndexWithExtent(
                    axis=axis,
                    value=axis.iteration_symbol(),
                    extent=(offset[axis.to_idx()], offset[axis.to_idx()]),
                )
        grid_subset = dcir.GridSubset(intervals=intervals)
        return dcir.FieldAccessInfo(
            grid_subset=grid_subset,
            global_grid_subset=global_subset,
            dynamic_access=len(variable_offset_axes) > 0 or is_conditional or region is not None,
            variable_offset_axes=variable_offset_axes,
        )

    def visit_FieldAccess(
        self,
        node: oir.FieldAccess,
        *,
        he_grid,
        grid_subset,
        is_write: bool = False,
        is_conditional: bool = False,
        region=None,
        ctx: "AccessInfoCollector.Context",
        **kwargs,
    ):
        self.visit(
            node.offset,
            is_conditional=is_conditional,
            ctx=ctx,
            is_write=False,
            region=region,
            he_grid=he_grid,
            grid_subset=grid_subset,
            **kwargs,
        )

        if (is_write and not self.collect_write) or (not is_write and not self.collect_read):
            return

        access_info = self._make_access_info(
            node.offset,
            axes=ctx.axes[node.name],
            is_conditional=is_conditional,
            region=region,
            he_grid=he_grid,
            grid_subset=grid_subset,
        )
        ctx.access_infos[node.name] = access_info.union(
            ctx.access_infos.get(node.name, access_info)
        )


def compute_dcir_access_infos(
    oir_node,
    *,
    oir_decls=None,
    block_extents=None,
    collect_read=True,
    collect_write=True,
    include_full_domain=False,
    **kwargs,
) -> dace.properties.DictProperty:
    if block_extents is None:
        assert isinstance(oir_node, oir.Stencil)
        block_extents = compute_horizontal_block_extents(oir_node)

    axes = {
        name: axes_list_from_flags(decl.dimensions)
        for name, decl in oir_decls.items()
        if isinstance(decl, oir.FieldDecl)
    }
    ctx = AccessInfoCollector.Context(axes=axes, access_infos=dict())
    AccessInfoCollector(collect_read=collect_read, collect_write=collect_write).visit(
        oir_node, block_extents=block_extents, ctx=ctx, **kwargs
    )
    if include_full_domain:
        res = dict()
        for name, access_info in ctx.access_infos.items():
            res[name] = access_info.union(
                dcir.FieldAccessInfo(
                    grid_subset=dcir.GridSubset.full_domain(axes=access_info.axes()),
                    global_grid_subset=access_info.global_grid_subset,
                )
            )
    else:
        res = ctx.access_infos

    return res


def make_dace_subset(
    context_info: "dcir.FieldAccessInfo",
    access_info: "dcir.FieldAccessInfo",
    data_dims: Tuple[int, ...],
) -> dace.subsets.Range:
    clamped_access_info = access_info
    clamped_context_info = context_info
    for axis in access_info.axes():
        if axis in access_info.variable_offset_axes:
            clamped_access_info = clamped_access_info.clamp_full_axis(axis)
        if axis in clamped_context_info.variable_offset_axes:
            clamped_context_info = clamped_context_info.clamp_full_axis(axis)
    res_ranges = []

    for axis in clamped_access_info.axes():
        context_start, _ = clamped_context_info.grid_subset.intervals[axis].to_dace_symbolic()
        subset_start, subset_end = clamped_access_info.grid_subset.intervals[
            axis
        ].to_dace_symbolic()
        res_ranges.append((subset_start - context_start, subset_end - context_start - 1, 1))
    res_ranges.extend((0, dim - 1, 1) for dim in data_dims)
    return dace.subsets.Range(res_ranges)


def untile_memlets(
    memlets: Sequence["dcir.Memlet"], axes: Sequence["dcir.Axis"]
) -> List["dcir.Memlet"]:
    res_memlets: List["dcir.Memlet"] = []
    for memlet in memlets:
        res_memlets.append(
            dcir.Memlet(
                field=memlet.field,
                access_info=memlet.access_info.untile(axes),
                connector=memlet.connector,
                is_read=memlet.is_read,
                is_write=memlet.is_write,
            )
        )
    return res_memlets


def union_node_grid_subsets(nodes: List[eve.Node]):
    grid_subset = None

    for node in collect_toplevel_iteration_nodes(nodes):
        if grid_subset is None:
            grid_subset = node.grid_subset
        grid_subset = grid_subset.union(node.grid_subset)

    return grid_subset


def _union_memlets(*memlets: "dcir.Memlet") -> List["dcir.Memlet"]:
    res: Dict[str, dcir.Memlet] = {}
    for memlet in memlets:
        res[memlet.field] = memlet.union(res.get(memlet.field, memlet))
    return list(res.values())


def union_inout_memlets(nodes: List[eve.Node]):
    read_memlets: List[dcir.Memlet] = []
    write_memlets: List[dcir.Memlet] = []
    for node in collect_toplevel_computation_nodes(nodes):
        read_memlets = _union_memlets(*read_memlets, *node.read_memlets)
        write_memlets = _union_memlets(*write_memlets, *node.write_memlets)

    return (read_memlets, write_memlets, _union_memlets(*read_memlets, *write_memlets))


def flatten_list(list_or_node: Union[List[Any], eve.Node]):
    list_or_node = [list_or_node]
    while not all(isinstance(ref, eve.Node) for ref in list_or_node):
        list_or_node = [r for li in list_or_node for r in li]
    return list_or_node


def collect_toplevel_computation_nodes(
    list_or_node: Union[List[Any], eve.Node]
) -> List["dcir.ComputationNode"]:
    class ComputationNodeCollector(eve.NodeVisitor):
        def visit_ComputationNode(self, node: dcir.ComputationNode, *, collection: List):
            collection.append(node)

    collection: List[dcir.ComputationNode] = []
    ComputationNodeCollector().visit(list_or_node, collection=collection)
    return collection


def collect_toplevel_iteration_nodes(
    list_or_node: Union[List[Any], eve.Node]
) -> List["dcir.IterationNode"]:
    class IterationNodeCollector(eve.NodeVisitor):
        def visit_IterationNode(self, node: dcir.IterationNode, *, collection: List):
            collection.append(node)

    collection: List[dcir.IterationNode] = []
    IterationNodeCollector().visit(list_or_node, collection=collection)
    return collection



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import collections
from typing import Any, Callable, Dict, Set, Union

from gt4py import eve
from gt4py.cartesian.gtc import oir

from .utils import AccessCollector, collect_symbol_names, symbol_name_creator


class TemporariesToScalarsBase(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    def visit_FieldAccess(
        self, node: oir.FieldAccess, *, tmps_name_map: Dict[str, str], **kwargs: Any
    ) -> Union[oir.FieldAccess, oir.ScalarAccess]:
        offsets = node.offset.to_dict()
        if node.name in tmps_name_map:
            assert (
                offsets["i"] == offsets["j"] == offsets["k"] == 0
            ), "Non-zero offset in temporary that is replaced?!"
            return oir.ScalarAccess(name=tmps_name_map[node.name], dtype=node.dtype)
        return self.generic_visit(node, tmps_name_map=tmps_name_map, **kwargs)

    def visit_HorizontalExecution(
        self,
        node: oir.HorizontalExecution,
        tmps_to_replace: Set[str],
        symtable: Dict[str, Any],
        new_symbol_name: Callable[[str], str],
        **kwargs: Any,
    ) -> oir.HorizontalExecution:
        local_tmps_to_replace = (
            node.walk_values()
            .if_isinstance(oir.FieldAccess)
            .getattr("name")
            .if_in(tmps_to_replace)
            .to_set()
        )
        tmps_name_map = {tmp: new_symbol_name(tmp) for tmp in local_tmps_to_replace}

        return oir.HorizontalExecution(
            body=self.visit(node.body, tmps_name_map=tmps_name_map, symtable=symtable, **kwargs),
            declarations=node.declarations
            + [
                oir.LocalScalar(
                    name=tmps_name_map[tmp], dtype=symtable[tmp].dtype, loc=symtable[tmp].loc
                )
                for tmp in local_tmps_to_replace
            ],
            loc=node.loc,
        )

    def visit_VerticalLoop(
        self,
        node: oir.VerticalLoop,
        tmps_to_replace: Set[str],
        **kwargs: Any,
    ) -> oir.VerticalLoop:
        return oir.VerticalLoop(
            loop_order=node.loop_order,
            sections=self.visit(node.sections, tmps_to_replace=tmps_to_replace, **kwargs),
            caches=[c for c in node.caches if c.name not in tmps_to_replace],
            loc=node.loc,
        )

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        tmps_to_replace = kwargs["tmps_to_replace"]
        all_names = collect_symbol_names(node)
        return oir.Stencil(
            name=node.name,
            params=node.params,
            vertical_loops=self.visit(
                node.vertical_loops,
                new_symbol_name=symbol_name_creator(all_names),
                **kwargs,
            ),
            declarations=[d for d in node.declarations if d.name not in tmps_to_replace],
            loc=node.loc,
        )


class LocalTemporariesToScalars(TemporariesToScalarsBase):
    """Replaces temporary fields accessed only within a single horizontal execution by scalars.

    1. Finds temporaries that are only accessed within a single HorizontalExecution.
    2. Replaces corresponding FieldAccess nodes by ScalarAccess nodes.
    3. Removes matching temporaries from VerticalLoop declarations.
    4. Add matching temporaries to HorizontalExecution declarations.

    Note that temporaries used in horizontal regions in a single horizontal execution
    may not be scalarized.

    """

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        horizontal_executions = node.walk_values().if_isinstance(oir.HorizontalExecution)
        temps_without_data_dims = set(
            [decl.name for decl in node.declarations if not decl.data_dims]
        )
        counts: collections.Counter = sum(
            (
                collections.Counter(
                    horizontal_execution.walk_values()
                    .if_isinstance(oir.FieldAccess)
                    .getattr("name")
                    .if_in(temps_without_data_dims)
                    .to_set()
                )
                for horizontal_execution in horizontal_executions
            ),
            collections.Counter(),
        )

        local_tmps = {tmp for tmp, count in counts.items() if count == 1}
        return super().visit_Stencil(node, tmps_to_replace=local_tmps, **kwargs)


class WriteBeforeReadTemporariesToScalars(TemporariesToScalarsBase):
    """Replaces temporay fields that are always written before read by scalars.

    Note that temporaries used in horizontal regions in a single horizontal execution
    may not be scalarized.

    """

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        # Does not (yet) support scalarizing temporaries with data_dims
        write_before_read_tmps = {
            symbol
            for symbol, value in kwargs["symtable"].items()
            if isinstance(value, oir.Temporary) and not value.data_dims
        }
        horizontal_executions = node.walk_values().if_isinstance(oir.HorizontalExecution)

        for horizontal_execution in horizontal_executions:
            accesses = AccessCollector.apply(horizontal_execution)
            offsets = accesses.offsets()
            ordered_accesses = accesses.ordered_accesses()

            def write_before_read(
                tmp: str, offsets=offsets, ordered_accesses=ordered_accesses
            ) -> bool:
                if tmp not in offsets:
                    return True
                if offsets[tmp] != {(0, 0, 0)}:
                    return False
                return next(
                    o.is_write and o.horizontal_mask is None
                    for o in ordered_accesses
                    if o.field == tmp
                )

            write_before_read_tmps = {
                tmp for tmp in write_before_read_tmps if write_before_read(tmp)
            }

        return super().visit_Stencil(node, tmps_to_replace=write_before_read_tmps, **kwargs)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import base64
import pickle
import typing
from typing import Dict, List, Set, Union

import dace.data
import dace.dtypes
import dace.properties
import dace.subsets
import numpy as np
from dace import library

from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc import daceir as dcir
from gt4py.cartesian.gtc import oir
from gt4py.cartesian.gtc.dace.expansion.expansion import StencilComputationExpansion
from gt4py.cartesian.gtc.definitions import Extent
from gt4py.cartesian.gtc.oir import Decl, FieldDecl, VerticalLoop, VerticalLoopSection

from .expansion.utils import HorizontalExecutionSplitter, get_dace_debuginfo
from .expansion_specification import ExpansionItem, make_expansion_order


def _set_expansion_order(
    node: "StencilComputation", expansion_order: Union[List[ExpansionItem], List[str]]
):
    res = make_expansion_order(node, expansion_order)
    node._expansion_specification = res


def _set_tile_sizes_interpretation(node: "StencilComputation", tile_sizes_interpretation: str):
    valid_values = {"shape", "strides"}
    if tile_sizes_interpretation not in valid_values:
        raise ValueError(f"tile_sizes_interpretation must be one in {valid_values}.")
    node._tile_sizes_interpretation = tile_sizes_interpretation


class PickledProperty:
    def to_json(self, obj):
        protocol = pickle.DEFAULT_PROTOCOL
        pbytes = pickle.dumps(obj, protocol=protocol)
        jsonobj = dict(pickle=base64.b64encode(pbytes).decode("utf-8"))
        return jsonobj

    @classmethod
    def from_json(cls, d, sdfg=None):
        b64string = d["pickle"]
        byte_repr = base64.b64decode(b64string)
        return pickle.loads(byte_repr)


class PickledDataclassProperty(PickledProperty, dace.properties.DataclassProperty):
    pass


class PickledListProperty(PickledProperty, dace.properties.ListProperty):
    pass


class PickledDictProperty(PickledProperty, dace.properties.DictProperty):
    pass


@library.node
class StencilComputation(library.LibraryNode):
    implementations: Dict[str, dace.library.ExpandTransformation] = {
        "default": StencilComputationExpansion
    }
    default_implementation = "default"

    oir_node = PickledDataclassProperty(dtype=VerticalLoop, allow_none=True)

    declarations = PickledDictProperty(key_type=str, value_type=Decl, allow_none=True)
    extents = PickledDictProperty(key_type=int, value_type=Extent, allow_none=False)
    access_infos = PickledDictProperty(
        key_type=str, value_type=dcir.FieldAccessInfo, allow_none=True
    )

    device = dace.properties.EnumProperty(
        dtype=dace.DeviceType, default=dace.DeviceType.CPU, allow_none=True
    )
    expansion_specification = PickledListProperty(
        element_type=ExpansionItem,
        allow_none=True,
        setter=_set_expansion_order,
    )
    tile_sizes = PickledDictProperty(
        key_type=dcir.Axis,
        value_type=int,
        default={dcir.Axis.I: 8, dcir.Axis.J: 8, dcir.Axis.K: 8},
    )

    tile_sizes_interpretation = dace.properties.Property(
        setter=_set_tile_sizes_interpretation, dtype=str, default="strides"
    )

    symbol_mapping = dace.properties.DictProperty(
        key_type=str, value_type=dace.symbolic.pystr_to_symbolic, default=None, allow_none=True
    )
    _dace_library_name = "StencilComputation"

    def __init__(
        self,
        name="unnamed_vloop",
        oir_node: VerticalLoop = None,
        extents: Dict[int, Extent] = None,
        declarations: Dict[str, Decl] = None,
        expansion_order=None,
        *args,
        **kwargs,
    ):
        super().__init__(name=name, *args, **kwargs)

        from gt4py.cartesian.gtc.dace.utils import compute_dcir_access_infos

        if oir_node is not None:
            assert extents is not None
            assert declarations is not None
            extents_dict = dict()
            for i, section in enumerate(oir_node.sections):
                for j, he in enumerate(section.horizontal_executions):
                    extents_dict[j * len(oir_node.sections) + i] = extents[id(he)]

            self.oir_node = typing.cast(PickledDataclassProperty, oir_node)
            self.extents = extents_dict  # type: ignore
            self.declarations = declarations  # type: ignore
            self.symbol_mapping = {
                decl.name: dace.symbol(
                    decl.name,
                    dtype=dace.typeclass(np.dtype(common.data_type_to_typestr(decl.dtype)).type),
                )
                for decl in declarations.values()
                if isinstance(decl, oir.ScalarDecl)
            }
            self.symbol_mapping.update(
                {
                    axis.domain_symbol(): dace.symbol(axis.domain_symbol(), dtype=dace.int32)
                    for axis in dcir.Axis.dims_horizontal()
                }
            )
            self.access_infos = compute_dcir_access_infos(
                oir_node,
                oir_decls=declarations,
                block_extents=self.get_extents,
                collect_read=True,
                collect_write=True,
            )
            if any(
                interval.start.level == common.LevelMarker.END
                or interval.end.level == common.LevelMarker.END
                for interval in oir_node.walk_values()
                .if_isinstance(VerticalLoopSection)
                .getattr("interval")
            ) or any(
                decl.dimensions[dcir.Axis.K.to_idx()]
                for decl in self.declarations.values()
                if isinstance(decl, oir.FieldDecl)
            ):
                self.symbol_mapping[dcir.Axis.K.domain_symbol()] = dace.symbol(
                    dcir.Axis.K.domain_symbol(), dtype=dace.int32
                )

            self.debuginfo = get_dace_debuginfo(oir_node)

        if expansion_order is None:
            expansion_order = [
                "TileI",
                "TileJ",
                "Sections",
                "K",  # Expands to either Loop or Map
                "Stages",
                "I",
                "J",
            ]
        _set_expansion_order(self, expansion_order)

    def get_extents(self, he):
        for i, section in enumerate(self.oir_node.sections):
            for j, cand_he in enumerate(section.horizontal_executions):
                if he is cand_he:
                    return self.extents[j * len(self.oir_node.sections) + i]

    @property
    def field_decls(self) -> Dict[str, FieldDecl]:
        return {
            name: decl for name, decl in self.declarations.items() if isinstance(decl, FieldDecl)
        }

    @property
    def free_symbols(self) -> Set[str]:
        result: Set[str] = set()
        for v in self.symbol_mapping.values():
            result.update(map(str, v.free_symbols))
        return result

    def has_splittable_regions(self):
        for he in self.oir_node.walk_values().if_isinstance(oir.HorizontalExecution):
            if not HorizontalExecutionSplitter.is_horizontal_execution_splittable(he):
                return False
        return True

    @property
    def tile_strides(self):
        if self.tile_sizes_interpretation == "strides":
            return self.tile_sizes
        else:
            overall_extent: Extent = next(iter(self.extents.values()))
            for extent in self.extents.values():
                overall_extent |= extent
            return {
                key: value + overall_extent[key.to_idx()] for key, value in self.tile_sizes.items()
            }



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tools for source code generation."""


from __future__ import annotations

import abc
import collections.abc
import contextlib
import inspect
import os
import re
import string
import subprocess
import sys
import textwrap
import types

import black
import jinja2
from mako import template as mako_tpl  # type: ignore[import]

from . import exceptions, utils
from .concepts import CollectionNode, LeafNode, Node, RootNode
from .extended_typing import (
    Any,
    Callable,
    ClassVar,
    Collection,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
    runtime_checkable,
)
from .visitors import NodeVisitor


SourceFormatter = Callable[[str], str]

SOURCE_FORMATTERS: Dict[str, SourceFormatter] = {}
"""Global dict storing registered formatters."""


class FormatterNameError(exceptions.EveRuntimeError):
    """Run-time error registering a new source code formatter."""

    ...


class FormattingError(exceptions.EveRuntimeError):
    """Run-time error applying a source code formatter."""

    ...


class TemplateDefinitionError(exceptions.EveTypeError):
    """Template definition error."""

    ...


class TemplateRenderingError(exceptions.EveRuntimeError):
    """Run-time error rendering a template."""

    ...


def register_formatter(language: str) -> Callable[[SourceFormatter], SourceFormatter]:
    """Register source code formatters for specific languages (decorator)."""

    def _decorator(formatter: SourceFormatter) -> SourceFormatter:
        if language in SOURCE_FORMATTERS:
            raise FormatterNameError(f"Another formatter for language '{language}' already exists")

        assert callable(formatter)
        SOURCE_FORMATTERS[language] = formatter

        return formatter

    return _decorator


@register_formatter("python")
def format_python_source(
    source: str,
    *,
    line_length: int = 100,
    python_versions: Optional[Set[str]] = None,
    string_normalization: bool = True,
) -> str:
    """Format Python source code using black formatter."""
    python_versions = python_versions or {f"{sys.version_info.major}{sys.version_info.minor}"}
    target_versions = set(
        black.TargetVersion[f"PY{v.replace('.', '')}"]  # type: ignore[attr-defined]
        for v in python_versions
    )

    formatted_source = black.format_str(
        source,
        mode=black.FileMode(
            line_length=line_length,
            target_versions=target_versions,
            string_normalization=string_normalization,
        ),
    )
    assert isinstance(formatted_source, str)

    return formatted_source


def _get_clang_format() -> Optional[str]:
    """Return the clang-format executable, or None if not available."""
    executable = os.getenv("CLANG_FORMAT_EXECUTABLE", "clang-format")
    try:
        assert isinstance(executable, str)
        if subprocess.run([executable, "--version"], capture_output=True).returncode != 0:
            return None
    except Exception:
        return None

    return executable


_CLANG_FORMAT_EXECUTABLE = _get_clang_format()


if _CLANG_FORMAT_EXECUTABLE is not None:

    @register_formatter("cpp")
    def format_cpp_source(
        source: str,
        *,
        style: Optional[str] = None,
        fallback_style: Optional[str] = None,
        sort_includes: bool = False,
    ) -> str:
        """Format C++ source code using clang-format."""
        assert isinstance(_CLANG_FORMAT_EXECUTABLE, str)
        args = [_CLANG_FORMAT_EXECUTABLE]
        if style:
            args.append(f"--style={style}")
        if fallback_style:
            args.append(f"--fallback-style={style}")
        if sort_includes:
            args.append("--sort-includes")

        try:
            # use a timeout as clang-format used to deadlock on some sources
            formatted_source = subprocess.run(
                args, check=True, input=source, capture_output=True, text=True, timeout=3
            ).stdout
        except subprocess.TimeoutExpired:
            return source

        assert isinstance(formatted_source, str)
        return formatted_source


def format_source(language: str, source: str, *, skip_errors: bool = True, **kwargs: Any) -> str:
    """Format source code if a formatter exists for the specific language."""
    formatter = SOURCE_FORMATTERS.get(language, None)
    try:
        if formatter:
            return formatter(source, **kwargs)
        else:
            raise FormattingError(f"Missing formatter for '{language}' language")
    except Exception as e:
        if skip_errors:
            return source
        else:
            raise FormattingError(
                f"Something went wrong when trying to format '{language}' source code"
            ) from e


class Name:
    """Text formatter with different case styles for symbol names in source code."""

    words: List[str]

    @classmethod
    def from_string(cls, name: str, case_style: utils.CaseStyleConverter.CASE_STYLE) -> Name:
        return cls(utils.CaseStyleConverter.split(name, case_style))

    def __init__(self, words: utils.AnyWordsIterable) -> None:
        if isinstance(words, str):
            words = [words]
        if not isinstance(words, collections.abc.Iterable):
            raise TypeError(
                f"Identifier definition ('{words}') type is not a valid sequence of words"
            )

        words = [*words]
        if not all(isinstance(item, str) for item in words):
            raise TypeError(
                f"Identifier definition ('{words}') type is not a valid sequence of words"
            )

        self.words = words

    def as_case(self, case_style: utils.CaseStyleConverter.CASE_STYLE) -> str:
        return utils.CaseStyleConverter.join(self.words, case_style)


AnyTextSequence = Union[Sequence[str], "TextBlock"]


class TextBlock:
    """A block of source code represented as a sequence of text lines.

    Check the provided context manager creator method (:meth:`indented`)
    for simple `indent - append - dedent` workflows.

    Args:
        indent_level: Initial indentation level.
        indent_size: Number of characters per indentation level.
        indent_char: Character used in the indentation.
        end_line: Character or string used as new-line separator.

    """

    def __init__(
        self,
        *,
        indent_level: int = 0,
        indent_size: int = 4,
        indent_char: str = " ",
        end_line: str = "\n",
    ) -> None:
        if not isinstance(indent_char, str) or len(indent_char) != 1:
            raise ValueError("'indent_char' must be a single-character string")
        if not isinstance(end_line, str):
            raise ValueError("'end_line' must be a string")

        self.indent_level = indent_level
        self.indent_size = indent_size
        self.indent_char = indent_char
        self.end_line = end_line
        self.lines: List[str] = []

    def append(self, new_line: str, *, update_indent: int = 0) -> TextBlock:
        if update_indent > 0:
            self.indent(update_indent)
        elif update_indent < 0:
            self.dedent(-update_indent)

        self.lines.append(self.indent_str + new_line)

        return self

    def extend(self, new_lines: AnyTextSequence, *, dedent: bool = False) -> TextBlock:
        assert isinstance(new_lines, (collections.abc.Sequence, TextBlock))

        if dedent:
            if isinstance(new_lines, TextBlock):
                new_lines = textwrap.dedent(new_lines.text).splitlines()
            else:
                new_lines = textwrap.dedent("\n".join(new_lines)).splitlines()

        elif isinstance(new_lines, TextBlock):
            new_lines = new_lines.lines

        for line in new_lines:
            self.append(line)

        return self

    def empty_line(self, count: int = 1) -> TextBlock:
        self.lines.extend([""] * count)
        return self

    def indent(self, steps: int = 1) -> TextBlock:
        self.indent_level += steps
        return self

    def dedent(self, steps: int = 1) -> TextBlock:
        assert self.indent_level >= steps
        self.indent_level -= steps
        return self

    @contextlib.contextmanager
    def indented(self, steps: int = 1) -> Iterator[TextBlock]:
        """Context manager creator for temporary indentation of sources.

        This context manager simplifies the usage of indent/dedent in
        common `indent - append - dedent` workflows.

        Examples:
            >>> block = TextBlock();
            >>> block.append('first line')  # doctest: +ELLIPSIS
            <...>
            >>> with block.indented():
            ...     block.append('second line');  # doctest: +ELLIPSIS
            <...>
            >>> block.append('third line')  # doctest: +ELLIPSIS
            <...>
            >>> print(block.text)
            first line
                second line
            third line

        """
        self.indent(steps)
        yield self
        self.dedent(steps)

    @property
    def text(self) -> str:
        """Single string with the whole block contents."""
        lines = ["".join([str(item) for item in line]) for line in self.lines]
        return self.end_line.join(lines)

    @property
    def indent_str(self) -> str:
        """Indentation string for new lines (in the current state)."""
        return self.indent_char * (self.indent_level * self.indent_size)

    def __iadd__(self, source_line: Union[str, AnyTextSequence]) -> TextBlock:
        if isinstance(source_line, str):
            return self.append(source_line)
        else:
            return self.extend(source_line)

    def __len__(self) -> int:
        return len(self.lines)

    def __str__(self) -> str:
        return self.text


TemplateT = TypeVar("TemplateT", bound="Template")


@runtime_checkable
class Template(Protocol):
    """Protocol (abstract base class) defining the Template interface.

    Direct subclassess of this base class only need to implement the
    abstract methods to adapt different template engines to this
    interface.

    Raises:
        TemplateDefinitionError: If the template definition contains errors.

    """

    def render(self, mapping: Optional[Mapping[str, str]] = None, **kwargs: Any) -> str:
        """Render the template.

        Args:
            mapping: A mapping whose keys match the template placeholders.

            ``**kwargs``: Placeholder values provided as keyword arguments (they will
                take precedence over ``mapping`` values for duplicated keys.

        Returns:
            The rendered string.

        Raises:
            TemplateRenderingError: If the template rendering fails.

        """
        if not mapping:
            mapping = {}
        if kwargs:
            mapping = {**mapping, **kwargs}

        return self.render_values(**mapping)

    @abc.abstractmethod
    def __init__(self, definition: Any, **kwargs: Any) -> None:
        pass

    @abc.abstractmethod
    def render_values(self, **kwargs: Any) -> str:
        """Render the template.

        Args:
            ``**kwargs``: Placeholder values provided as keyword arguments.

        Returns:
            The rendered string.

        Raises:
            TemplateRenderingError: If the template rendering fails.

        """
        pass


class BaseTemplate(Template):
    """Helper class to add source location info of template definitions."""

    definition: Any
    definition_loc: Optional[Tuple[str, int]]

    def __init__(self) -> None:
        self.definition_loc = None
        frame = inspect.currentframe()
        try:
            if frame is not None and frame.f_back is not None and frame.f_back.f_back is not None:
                (filename, lineno, _, _, _) = inspect.getframeinfo(frame.f_back.f_back)
                self.definition_loc = (filename, lineno)
        except Exception:
            pass
        finally:
            del frame

    def __str__(self) -> str:
        result = f"<{type(self).__qualname__}: '{self.definition}'>"
        if self.definition_loc:
            result += f" created at {self.definition_loc[0]}:{self.definition_loc[1]}"
        return result


class FormatTemplate(BaseTemplate):
    """Template adapter to render regular strings as fully-featured f-strings."""

    definition: str

    def __init__(self, definition: str, **kwargs: Any) -> None:
        super().__init__()
        self.definition = f'(f"""{definition}""")'

    def render_values(self, **kwargs: Any) -> str:
        try:
            result = eval(self.definition, {}, kwargs or {})
            assert isinstance(result, str)
            return result
        except Exception as e:
            message = f"<{type(self).__name__}: '{self.definition}'>"
            if self.definition_loc:
                message += f" (created at {self.definition_loc[0]}:{self.definition_loc[1]})"
            message += " rendering error."

            raise TemplateRenderingError(message, template=self) from e


class StringTemplate(BaseTemplate):
    """Template adapter for `string.Template`."""

    definition: string.Template

    def __init__(self, definition: Union[str, string.Template], **kwargs: Any) -> None:
        super().__init__()
        if isinstance(definition, str):
            definition = string.Template(definition)
        assert isinstance(definition, string.Template)
        self.definition = definition

    def render_values(self, **kwargs: Any) -> str:
        try:
            return self.definition.substitute(**kwargs)
        except Exception as e:
            message = f"<{type(self).__name__}>"
            if self.definition_loc:
                message += f" (created at {self.definition_loc[0]}:{self.definition_loc[1]})"
            try:
                loc_info = re.search(r"line (\d+), col (\d+)", str(e))
                message += f" rendering error at template line: {loc_info[1]}, column {loc_info[2]}."  # type: ignore
            except Exception:
                message += " rendering error."

            raise TemplateRenderingError(message, template=self) from e


class JinjaTemplate(BaseTemplate):
    """Template adapter for `jinja2.Template`."""

    definition: jinja2.Template

    __jinja_env__ = jinja2.Environment(undefined=jinja2.StrictUndefined)

    def __init__(self, definition: Union[str, jinja2.Template], **kwargs: Any) -> None:
        super().__init__()
        try:
            if isinstance(definition, str):
                definition = self.__jinja_env__.from_string(definition)
            assert isinstance(definition, jinja2.Template)
            self.definition = definition
        except Exception as e:
            message = "Error in JinjaTemplate"
            if self.definition_loc:
                message += f" created at {self.definition_loc[0]}:{self.definition_loc[1]}"
                try:
                    if hasattr(e, "lineno"):
                        message += f" (error likely around line: {e.lineno})"  # type: ignore  # assume Jinja exception
                except Exception:
                    message = f"{message}:\n---\n{definition}\n---\n"

            raise TemplateDefinitionError(message, definition=definition) from e

    def render_values(self, **kwargs: Any) -> str:
        try:
            return self.definition.render(**kwargs)
        except Exception as e:
            message = f"<{type(self).__name__}>"
            if self.definition_loc:
                message += f" (created at {self.definition_loc[0]}:{self.definition_loc[1]})"
            try:
                message += f" rendering error at template line: {e.lineno}."  # type: ignore  # assume Jinja exception
            except Exception:
                message += " rendering error."

            raise TemplateRenderingError(message, template=self) from e


class MakoTemplate(BaseTemplate):
    """Template adapter for `mako.template.Template`."""

    definition: mako_tpl.Template

    def __init__(self, definition: mako_tpl.Template, **kwargs: Any) -> None:
        super().__init__()
        try:
            if isinstance(definition, str):
                definition = mako_tpl.Template(definition)
            assert isinstance(definition, mako_tpl.Template)
            self.definition = definition
        except Exception as e:
            message = "Error in MakoTemplate"
            if self.definition_loc:
                message += f" created at {self.definition_loc[0]}:{self.definition_loc[1]}"
                try:
                    message += f" (error likely around line {e.lineno}, column: {getattr(e, 'pos', '?')})"  # type: ignore  # assume Mako exception
                except Exception:
                    message = f"{message}:\n---\n{definition}\n---\n"

            raise TemplateDefinitionError(message, definition=definition) from e

    def render_values(self, **kwargs: Any) -> str:
        try:
            result = self.definition.render(**kwargs)
            assert isinstance(result, str)
            return result
        except Exception as e:
            message = f"<{type(self).__name__}>"
            if self.definition_loc:
                message += f" (created at {self.definition_loc[0]}:{self.definition_loc[1]})"
            try:
                message += f" rendering error at template line: {e.lineno}, column: {getattr(e, 'pos', '?')}"  # type: ignore  # assume Mako exception
            except Exception:
                message += " rendering error."
            raise TemplateRenderingError(message, template=self) from e


class TemplatedGenerator(NodeVisitor):
    """A code generator visitor using :class:`TextTemplate`.

    The order followed to choose a `dump()` function for node values is the following:

        1. A ``self.visit_NODE_CLASS_NAME()`` method where `NODE_CLASS_NAME`
           matches ``type(node).__name__``.
        2. A ``self.visit_NODE_BASE_CLASS_NAME()`` method where
           `NODE_BASE_CLASS_NAME` matches ``base.__name__``, and `base` is
           one of the node base classes (evaluated following the order
           given in ``type(node).__mro__``).

        If `node` is an instance of :class:`eve.Node` and a `visit_` method has
        not been found, `TemplatedGenerator` will look for an appropriate
        :class:`Template` definition:

        3. A ``NODE_CLASS_NAME`` class variable of type :class:`Template`,
           where ``NODE_CLASS_NAME`` matches ``type(node).__name__``.
        4. A ``NODE_BASE_CLASS_NAME`` class variable of type :class:`Template`,
           where ``NODE_BASE_CLASS_NAME`` matches ``base.__name__``, and
           `base` is one of the node base classes (evaluated following the
           order given in ``type(node).__mro__``).

        In any other case (templates cannot be used for instances of arbitrary types),
        steps 3 and 4 will be substituted by a call to the :meth:`self.generic_dump()`
        method.

    The following keys are passed to template instances at rendering:

        * ``**node_fields``: all the node children and implementation fields by name.
        * ``_impl``: a ``dict`` instance with the results of visiting all
          the node implementation fields.
        * ``_children``: a ``dict`` instance with the results of visiting all
          the node children.
        * ``_this_node``: the actual node instance (before visiting children).
        * ``_this_generator``: the current generator instance.
        * ``_this_module``: the generator's module instance.
        * ``**kwargs``: the keyword arguments received by the visiting method.

    Visitor methods can trigger regular template rendering for the same node class
    by explicitly calling :meth:`generic_visit()` (typically at the end), which will
    continue the search rendering algorithm at step 3. Thus, a common pattern to deal
    with complicated nodes is to define both a visitor method and a template for
    the same class, where the visitor method preprocess the node data and calls
    :meth:`generic_visit()` at the end with additional keyword arguments which will
    be forwarded to the node template.

    """

    __templates__: ClassVar[Mapping[str, Template]]

    @classmethod
    def __init_subclass__(cls, *, inherit_templates: bool = True, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if "__templates__" in cls.__dict__:
            raise TypeError(f"Invalid '__templates__' member in class {cls}")

        templates: Dict[str, Template] = {}
        if inherit_templates:
            for templated_gen_class in reversed(cls.__mro__[1:]):
                if (
                    issubclass(templated_gen_class, TemplatedGenerator)
                    and templated_gen_class is not TemplatedGenerator
                ):
                    templates.update(templated_gen_class.__templates__)

        templates.update(
            {
                key: value
                for key, value in cls.__dict__.items()
                if isinstance(value, Template) and not key.startswith("_") and not key.endswith("_")
            }
        )

        cls.__templates__ = types.MappingProxyType(templates)

    @overload
    @classmethod
    def apply(cls, root: LeafNode, **kwargs: Any) -> str:
        ...

    @overload
    @classmethod
    def apply(  # noqa: F811  # redefinition of symbol
        cls, root: CollectionNode, **kwargs: Any
    ) -> Collection[str]:
        ...

    @classmethod
    def apply(  # noqa: F811  # redefinition of symbol
        cls, root: RootNode, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        """Public method to build a class instance and visit an IR node.

        Args:
            root: An IR node.
            node_templates (optiona): see :class:`NodeDumper`.
            dump_function (optiona): see :class:`NodeDumper`.
            ``**kwargs`` (optional): custom extra parameters forwarded to `visit_NODE_TYPE_NAME()`.

        Returns:
            String (or collection of strings) with the dumped version of the root IR node.

        """
        return cls().visit(root, **kwargs)

    @classmethod
    def generic_dump(cls, node: RootNode, **kwargs: Any) -> str:
        """Class-specific ``dump()`` function for primitive types.

        This class could be redefined in the subclasses.
        """
        return str(node)

    def generic_visit(self, node: RootNode, **kwargs: Any) -> Union[str, Collection[str]]:
        if isinstance(node, Node):
            template, key = self.get_template(node)
            if template:
                try:
                    return self.render_template(
                        template,
                        node,
                        self.transform_children(node, **kwargs),
                        self.transform_annexed_items(node, **kwargs),
                        **kwargs,
                    )
                except TemplateRenderingError as e:
                    # Raise a new exception with extra information keeping the original cause
                    raise TemplateRenderingError(
                        f"Error in '{key}' template when rendering node '{node}'.\n"
                        + getattr(e, "message", str(e)),
                        **e.info,
                        node=node,
                    ) from e.__cause__

        if isinstance(node, (list, tuple, collections.abc.Set)) or (
            isinstance(node, collections.abc.Sequence) and not isinstance(node, (str, bytes))
        ):
            return [self.visit(value, **kwargs) for value in node]
        elif isinstance(node, (dict, collections.abc.Mapping)):
            return {key: self.visit(value, **kwargs) for key, value in node.items()}

        return self.generic_dump(node, **kwargs)

    def get_template(self, node: RootNode) -> Tuple[Optional[Template], Optional[str]]:
        """Get a template for a node instance (see class documentation)."""
        template: Optional[Template] = None
        template_key = None
        if isinstance(node, Node):
            for node_class in node.__class__.__mro__:
                template_key = node_class.__name__
                template = self.__templates__.get(template_key, None)
                if template is not None or node_class is Node:
                    break

        return template, None if template is None else template_key

    def render_template(
        self,
        template: Template,
        node: Node,
        transformed_children: Mapping[str, Any],
        transformed_annexed_items: Mapping[str, Any],
        **kwargs: Any,
    ) -> str:
        """Render a template using node instance data (see class documentation)."""
        return template.render(
            **{**transformed_children, **transformed_annexed_items, **kwargs},
            _children=transformed_children,
            _impl=transformed_annexed_items,
            _this_node=node,
            _this_generator=self,
            _this_module=sys.modules[type(self).__module__],
        )

    def transform_children(self, node: Node, **kwargs: Any) -> Dict[str, Any]:
        return {key: self.visit(value, **kwargs) for key, value in node.iter_children_items()}  # type: ignore[misc]

    def transform_annexed_items(self, node: Node, **kwargs: Any) -> Dict[str, Any]:
        return {key: self.visit(value, **kwargs) for key, value in node.annex.items()}



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numbers
import textwrap
from dataclasses import dataclass, field
from typing import Any, Collection, List, Optional, Set, Tuple, Union, cast

from gt4py import eve
from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc.numpy import npir
from gt4py.eve import codegen
from gt4py.eve.codegen import FormatTemplate as as_fmt
from gt4py.eve.codegen import JinjaTemplate as as_jinja


__all__ = ["NpirCodegen"]


def _offset_to_str(offset: int) -> str:
    if offset > 0:
        return f" + {offset}"
    elif offset < 0:
        return f" - {-offset}"
    else:
        return ""


def _slice_string(ch: str, offset: int, interval: Tuple[common.AxisBound, common.AxisBound]) -> str:
    start_ch = ch if interval[0].level == common.LevelMarker.START else ch.upper()
    end_ch = ch if interval[1].level == common.LevelMarker.START else ch.upper()

    return (
        f"{start_ch}{_offset_to_str(interval[0].offset + offset)}"
        ":"
        f"{end_ch}{_offset_to_str(interval[1].offset + offset)}"
    )


def _make_slice_access(
    offset: Tuple[Optional[int], Optional[int], Union[str, Optional[int]]],
    is_serial: bool,
    interval: Optional[npir.HorizontalMask] = None,
) -> List[str]:
    axes: List[str] = []

    if interval is None:
        interval = npir.HorizontalMask(
            i=(common.AxisBound.start(0), common.AxisBound.end(0)),
            j=(common.AxisBound.start(0), common.AxisBound.end(0)),
        )

    if offset[0] is not None:
        axes.append(_slice_string("i", offset[0], interval.i))
    if offset[1] is not None:
        axes.append(_slice_string("j", offset[1], interval.j))

    if isinstance(offset[2], numbers.Number):
        bounds = (
            (common.AxisBound.start(), common.AxisBound.start(offset=1))
            if is_serial
            else (common.AxisBound.start(), common.AxisBound.end())
        )
        k_str = "k_" if is_serial else "k"
        axes.append(_slice_string(k_str, offset[2], bounds))
    elif isinstance(offset[2], str):
        axes.append(offset[2])

    return axes


ORIGIN_CORRECTED_VIEW_CLASS = textwrap.dedent(
    """\
    class Field:
        def __init__(self, field, offsets: Tuple[int, ...], dimensions: Tuple[bool, bool, bool]):
            ii = iter(range(3))
            self.idx_to_data = tuple(
                [next(ii) if has_dim else None for has_dim in dimensions]
                + list(range(sum(dimensions), len(field.shape)))
            )

            shape = [field.shape[i] if i is not None else 1 for i in self.idx_to_data]
            self.field_view = np.reshape(field.data, shape).view(np.ndarray)

            self.offsets = offsets

        @classmethod
        def empty(cls, shape, dtype, offset):
            return cls(np.empty(shape, dtype=dtype), offset, (True, True, True))

        def shim_key(self, key):
            new_args = []
            if not isinstance(key, tuple):
                key = (key, )
            for index in self.idx_to_data:
                if index is None:
                    new_args.append(slice(None, None))
                else:
                    idx = key[index]
                    offset = self.offsets[index]
                    if isinstance(idx, slice):
                        new_args.append(
                            slice(idx.start + offset, idx.stop + offset, idx.step) if offset else idx
                        )
                    else:
                        new_args.append(idx + offset)
            if not isinstance(new_args[2], (numbers.Integral, slice)):
                new_args = self.broadcast_and_clip_variable_k(new_args)
            return tuple(new_args)

        def broadcast_and_clip_variable_k(self, new_args: tuple):
            assert isinstance(new_args[0], slice) and isinstance(new_args[1], slice)
            if np.max(new_args[2]) >= self.field_view.shape[2] or np.min(new_args[2]) < 0:
                new_args[2] = np.clip(new_args[2].copy(), 0, self.field_view.shape[2]-1)
            new_args[:2] = np.broadcast_arrays(
                np.expand_dims(
                    np.arange(new_args[0].start, new_args[0].stop),
                    axis=tuple(i for i in range(self.field_view.ndim) if i != 0)
                ),
                np.expand_dims(
                    np.arange(new_args[1].start, new_args[1].stop),
                    axis=tuple(i for i in range(self.field_view.ndim) if i != 1)
                ),
            )
            return new_args

        def __getitem__(self, key):
            return self.field_view.__getitem__(self.shim_key(key))

        def __setitem__(self, key, value):
            return self.field_view.__setitem__(self.shim_key(key), value)
    """
)


class NpirCodegen(codegen.TemplatedGenerator, eve.VisitorWithSymbolTableTrait):
    @dataclass
    class BlockContext:
        locals_declared: Set[str] = field(default_factory=set)

        def add_declared(self, *args):
            self.locals_declared |= set(args)

    FieldDecl = as_fmt("{name} = Field({name}, _origin_['{name}'], ({', '.join(dimensions)}))")

    def visit_TemporaryDecl(
        self, node: npir.TemporaryDecl, **kwargs
    ) -> Union[str, Collection[str]]:
        shape = [f"_dI_ + {node.padding[0]}", f"_dJ_ + {node.padding[1]}", "_dK_"] + [
            str(dim) for dim in node.data_dims
        ]
        offset = [str(off) for off in node.offset] + ["0"] * (1 + len(node.data_dims))
        dtype = self.visit(node.dtype, **kwargs)
        return f"{node.name} = Field.empty(({', '.join(shape)}), {dtype}, ({', '.join(offset)}))"

    LocalScalarDecl = as_fmt(
        "{name} = Field.empty((_dI_ + {upper[0] + lower[0]}, _dJ_ + {upper[1] + lower[1]}, {ksize}), {dtype}, ({', '.join(str(l) for l in lower)}, 0))"
    )

    VarKOffset = as_fmt("lk + {k}")

    def visit_FieldSlice(self, node: npir.FieldSlice, **kwargs: Any) -> Union[str, Collection[str]]:
        k_offset = (
            self.visit(node.k_offset, **kwargs)
            if isinstance(node.k_offset, npir.VarKOffset)
            else node.k_offset
        )
        offsets: Tuple[Optional[int], Optional[int], Union[str, int, None]] = (
            node.i_offset,
            node.j_offset,
            k_offset,
        )

        # To determine: when is the symbol name not in the symtable?
        if node.name in kwargs.get("symtable", {}):
            decl = kwargs["symtable"][node.name]
            dimensions = decl.dimensions if isinstance(decl, npir.FieldDecl) else [True] * 3
            offsets = cast(
                Tuple[Optional[int], Optional[int], Union[str, int, None]],
                tuple(off if has_dim else None for has_dim, off in zip(dimensions, offsets)),
            )

        args = _make_slice_access(offsets, kwargs["is_serial"], kwargs.get("horizontal_mask"))
        data_index = self.visit(node.data_index, inside_slice=True, **kwargs)

        access_slice = ", ".join(args + list(data_index))

        return f"{node.name}[{access_slice}]"

    def visit_LocalScalarAccess(
        self,
        node: npir.LocalScalarAccess,
        *,
        is_serial: bool,
        horizontal_mask: Optional[npir.HorizontalMask] = None,
        **kwargs: Any,
    ) -> Union[str, Collection[str]]:
        args = _make_slice_access((0, 0, 0), is_serial, horizontal_mask)
        if is_serial:
            args[2] = ":"
        return f"{node.name}[{', '.join(args)}]"

    ParamAccess = as_fmt("{name}")

    def visit_DataType(self, node: common.DataType, **kwargs: Any) -> Union[str, Collection[str]]:
        # `np.bool` is a deprecated alias for the builtin `bool` or `np.bool_`.
        if node not in {common.DataType.BOOL}:
            return f"np.{node.name.lower()}"
        else:
            return node.name.lower()

    def visit_BuiltInLiteral(
        self, node: common.BuiltInLiteral, **kwargs
    ) -> Union[str, Collection[str]]:
        if node is common.BuiltInLiteral.TRUE:
            return "True"
        elif node is common.BuiltInLiteral.FALSE:
            return "False"
        else:
            return self.generic_visit(node, **kwargs)

    def visit_ScalarLiteral(
        self, node: npir.ScalarLiteral, *, inside_slice: bool = False, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        # This could be trivial, but it's convenient for reading if the dtype is omitted in slices.
        dtype = self.visit(node.dtype, inside_slice=inside_slice, **kwargs)
        value = self.visit(node.value, inside_slice=inside_slice, **kwargs)
        return f"{value}" if inside_slice else f"{dtype}({value})"

    ScalarCast = as_fmt("{dtype}({expr})")

    VectorCast = as_fmt("{expr}.astype({dtype})")

    def visit_NativeFunction(
        self, node: common.NativeFunction, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        if node == common.NativeFunction.MIN:
            return "np.minimum"
        elif node == common.NativeFunction.MAX:
            return "np.maximum"
        elif node == common.NativeFunction.POW:
            return "np.power"
        elif node == common.NativeFunction.GAMMA:
            return "scipy.special.gamma"
        name = self.generic_visit(node, **kwargs)
        assert isinstance(name, str)
        return f"np.{name}"

    def visit_NativeFuncCall(
        self, node: npir.NativeFuncCall, *, mask: Optional[str] = None, **kwargs: Any
    ):
        kwargs["mask_arg"] = f", where={mask}" if mask else ""
        return self.generic_visit(node, mask=mask, **kwargs)

    NativeFuncCall = as_fmt("{func}({', '.join(arg for arg in args)}{mask_arg})")

    def visit_VectorAssign(
        self, node: npir.VectorAssign, *, ctx: "BlockContext", **kwargs: Any
    ) -> Union[str, Collection[str]]:
        left = self.visit(node.left, horizontal_mask=node.horizontal_mask, **kwargs)
        right = self.visit(node.right, horizontal_mask=node.horizontal_mask, **kwargs)
        return f"{left} = {right}"

    VectorArithmetic = as_fmt("({left} {op} {right})")

    VectorLogic = as_fmt("np.bitwise_{op}({left}, {right})")

    def visit_UnaryOperator(
        self, node: common.UnaryOperator, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        if node is common.UnaryOperator.NOT:
            return "np.bitwise_not"
        return self.generic_visit(node, **kwargs)

    VectorUnaryOp = as_fmt("({op}({expr}))")

    VectorTernaryOp = as_fmt("np.where({cond}, {true_expr}, {false_expr})")

    def visit_LevelMarker(
        self, node: common.LevelMarker, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        return "K" if node == common.LevelMarker.END else "k"

    def visit_AxisBound(self, node: common.AxisBound, **kwargs: Any) -> Union[str, Collection[str]]:
        if node.offset > 0:
            voffset = f" + {node.offset}"
        elif node.offset == 0:
            voffset = ""
        else:
            voffset = f" - {-node.offset}"
        return self.generic_visit(node, voffset=voffset, **kwargs)

    AxisBound = as_fmt("_d{level}_{voffset}")

    def visit_LoopOrder(self, node: common.LoopOrder, **kwargs) -> Union[str, Collection[str]]:
        if node is common.LoopOrder.FORWARD:
            return "for k_ in range(k, K):"
        elif node is common.LoopOrder.BACKWARD:
            return "for k_ in range(K-1, k-1, -1):"
        return ""

    Broadcast = as_fmt("{expr}")

    While = as_jinja(
        textwrap.dedent(
            """\
            while np.any({{ cond }}):
                {% for stmt in body %}{{ stmt }}
                {% endfor %}
            """
        )
    )

    def visit_While(self, node: npir.While, **kwargs: Any) -> str:
        cond = self.visit(node.cond, **kwargs)
        body = []
        for stmt in self.visit(node.body, **kwargs):
            body.extend(stmt.split("\n"))
        return self.While.render(cond=cond, body=body)

    def visit_VerticalPass(self, node: npir.VerticalPass, **kwargs):
        is_serial = node.direction != common.LoopOrder.PARALLEL
        has_variable_k = bool(node.walk_values().if_isinstance(npir.VarKOffset).to_list())
        return self.generic_visit(
            node,
            is_serial=is_serial,
            has_variable_k=has_variable_k,
            ksize="_dK_" if not is_serial else "1",
            lk_stmt="lk = " + ("k_" if is_serial else "np.arange(k, K)[np.newaxis, np.newaxis, :]"),
            **kwargs,
        )

    VerticalPass = as_jinja(
        textwrap.dedent(
            """\
            # --- begin vertical block ---{% set body_indent = 0 %}
            k, K = {{ lower }}, {{ upper }}
            {%- if direction %}
            {{ direction }}{% set body_indent = 4 %}{% endif %}
            {% if has_variable_k %}
            {{ lk_stmt | indent(body_indent, first=True) }}
            {% endif -%}
            {% for hblock in body %}
            {{ hblock | indent(body_indent, first=True) }}
            {% endfor %}# --- end vertical block ---
            """
        )
    )

    def visit_HorizontalBlock(
        self, node: npir.HorizontalBlock, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        lower = (-node.extent[0][0], -node.extent[1][0])
        upper = (node.extent[0][1], node.extent[1][1])
        return self.generic_visit(node, lower=lower, upper=upper, ctx=self.BlockContext(), **kwargs)

    HorizontalBlock = as_jinja(
        textwrap.dedent(
            """\
            # --- begin horizontal block --
            i, I = _di_ - {{ lower[0] }}, _dI_ + {{ upper[0] }}
            j, J = _dj_ - {{ lower[1] }}, _dJ_ + {{ upper[1] }}

            {% for stmt in body %}{{ stmt }}
            {% endfor -%}
            # --- end horizontal block --

            """
        )
    )

    def visit_Computation(
        self, node: npir.Computation, *, ignore_np_errstate: bool = True, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        signature = ["*", *node.arguments, "_domain_", "_origin_"]
        return self.generic_visit(
            node,
            signature=", ".join(signature),
            data_view_class=ORIGIN_CORRECTED_VIEW_CLASS,
            ignore_np_errstate=ignore_np_errstate,
            **kwargs,
        )

    Computation = as_jinja(
        textwrap.dedent(
            """\
            import numbers
            from typing import Tuple

            import numpy as np
            import scipy.special

            {{ data_view_class }}

            def run({{ signature }}):

                # --- begin domain boundary shortcuts ---
                _di_, _dj_, _dk_ = 0, 0, 0
                _dI_, _dJ_, _dK_ = _domain_
                # --- end domain padding ---

                {% for decl in api_field_decls %}{{ decl | indent(4) }}
                {% endfor %}
                {% for decl in temp_decls %}{{ decl | indent(4) }}
                {% endfor %}

                {% if ignore_np_errstate -%}
                with np.errstate(divide='ignore', over='ignore', under='ignore', invalid='ignore'):
                {%- else -%}
                with np.errstate():
                {%- endif %}

                {% for pass in vertical_passes %}
                {{ pass | indent(8) }}
                {% else %}
                    pass
                {% endfor %}
            """
        )
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Text and templating utilities."""

import collections.abc
import contextlib
import re
import textwrap

import black


black_mode = black.FileMode(
    target_versions={black.TargetVersion.PY36, black.TargetVersion.PY37}, line_length=120
)


def format_source(source: str, line_length: int) -> str:
    black_mode.line_length = line_length
    return black.format_str(source, mode=black_mode)


def get_line_number(text, re_query, re_flags=0):
    """Return (0-based) line number of the match or None."""
    prog = re.compile(re_query, re_flags)
    lines = text.splitlines()
    for n, line in enumerate(lines):
        if prog.search(line):
            return n

    return None


class Joiner:
    class Placeholder:
        def __init__(self, joiner, index):
            self.joiner = joiner
            self.index = index

        def __str__(self):
            return self.joiner.joiner_str if self.index < self.joiner.n_items - 1 else ""

    def __init__(self, joiner_str):
        self.joiner_str = joiner_str
        self.n_items = 0

    def __call__(self):
        self.n_items += 1
        return type(self).Placeholder(self, self.n_items - 1)


class TextBlock:
    def __init__(self, indent_level=0, indent_size=4, indent_char=" ", end_line="\n"):
        self.indent_level = indent_level
        self.indent_size = indent_size
        self.indent_char = indent_char
        self.end_line = end_line
        self.lines = []

    def append(self, source_line, indent_steps=0):
        self.indent(indent_steps)
        if isinstance(source_line, str):
            source_line = [source_line]

        line = [self.indent_char * (self.indent_level * self.indent_size)]
        for item in source_line:
            if isinstance(item, str) and isinstance(line[-1], str):
                line[-1] += item
            else:
                line.append(item)

        self.lines.append(line)

        return self

    def extend(self, source_lines, *, dedent=False):
        assert isinstance(source_lines, (str, collections.abc.Sequence, TextBlock))

        if dedent:
            if isinstance(source_lines, TextBlock):
                source_lines = source_lines.text
            elif not isinstance(source_lines, str):
                source_lines = "\n".join(source_lines)
            source_lines = textwrap.dedent(source_lines)

        if isinstance(source_lines, str):
            source_lines = source_lines.splitlines()
        elif isinstance(source_lines, TextBlock):
            source_lines = source_lines.lines

        for line in source_lines:
            self.append(line)
        return self

    def empty_line(self, steps=1):
        self.lines.append("")
        return self

    def indent(self, steps=1):
        self.indent_level += steps
        return self

    def dedent(self, steps=1):
        assert self.indent_level >= steps
        self.indent_level -= steps
        return self

    @contextlib.contextmanager
    def indented(self, steps=1):
        self.indent(steps)
        yield self
        self.dedent(steps)

    @property
    def text(self):
        lines = ["".join([str(item) for item in line]) for line in self.lines]
        return self.end_line.join(lines)

    def __iadd__(self, source_line):
        return self.append(source_line)

    def __len__(self):
        return len(self.lines)

    def __str__(self):
        return self.text



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Iterator utils."""


from __future__ import annotations

import abc
import collections.abc
import functools

from . import utils
from .extended_typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from .type_definitions import Enum


try:
    # For perfomance reasons, try to use cytoolz when possible (using cython)
    import cytoolz as toolz  # type: ignore[import]
except ModuleNotFoundError:
    # Fall back to pure Python toolz
    import toolz  # type: ignore[import] # noqa: F401  # imported but unused


TreeKey = Union[int, str]


if TYPE_CHECKING:
    TreeLike = Any
else:

    class TreeLike(abc.ABC):  # noqa: B024
        ...


TreeLikeT = TypeVar("TreeLikeT", bound=TreeLike)


class Tree(Protocol):
    @abc.abstractmethod
    def iter_children_values(self) -> Iterable:
        ...

    @abc.abstractmethod
    def iter_children_items(self) -> Iterable[Tuple[TreeKey, Any]]:
        ...


TreeLike.register(Tree)

_T = TypeVar("_T")


@functools.singledispatch
def iter_children_values(node: TreeLike) -> Iterable:
    """Create an iterator to traverse values as Eve tree nodes."""
    return node.iter_children_values() if hasattr(node, "iter_children_values") else iter(())


@functools.singledispatch
def iter_children_items(node: TreeLike) -> Iterable[Tuple[TreeKey, Any]]:
    """Create an iterator to traverse values as Eve tree nodes."""
    return node.iter_children_items() if hasattr(node, "iter_children_items") else iter(())


def register_tree_like(
    *types: Type[_T],
    iter_values_fn: Callable[[_T], Iterable],
    iter_items_fn: Callable[[_T], Iterable[Tuple[TreeKey, Any]]],
) -> None:
    for t in types:
        TreeLike.register(t)
        iter_children_values.register(t)(iter_values_fn)
        iter_children_items.register(t)(iter_items_fn)

    return None


register_tree_like(str, bytes, iter_values_fn=lambda _: iter(()), iter_items_fn=lambda _: iter(()))


register_tree_like(
    collections.abc.Sequence,  # type: ignore[misc]  # It should be concrete class
    collections.abc.Set,  # type: ignore[misc]  # It should be concrete class
    iter_values_fn=lambda x: iter(x),
    iter_items_fn=lambda x: enumerate(x),
)

register_tree_like(
    collections.abc.Mapping,  # type: ignore[misc]  # It should be concrete class
    iter_values_fn=lambda x: x.values(),
    iter_items_fn=lambda x: x.items(),
)


# -- Tree traversals --
class TraversalOrder(Enum):
    PRE_ORDER = "pre"
    POST_ORDER = "post"
    BFS_ORDER = "bfs"


def _pre_walk_items(
    node: TreeLike, *, __key__: Optional[TreeKey] = None
) -> Iterable[Tuple[Optional[TreeKey], Any]]:
    """Create a pre-order tree traversal iterator of (key, value) pairs."""
    yield __key__, node
    for key, child in iter_children_items(node):
        yield from _pre_walk_items(child, __key__=key)


def _pre_walk_values(node: TreeLike) -> Iterable[Tuple[Any]]:
    """Create a pre-order tree traversal iterator of values."""
    yield node
    for child in iter_children_values(node):
        yield from _pre_walk_values(child)


pre_walk_items = utils.as_xiter(_pre_walk_items)
pre_walk_values = utils.as_xiter(_pre_walk_values)


def _post_walk_items(
    node: TreeLike, *, __key__: Optional[TreeKey] = None
) -> Iterable[Tuple[Optional[TreeKey], Any]]:
    """Create a post-order tree traversal iterator of (key, value) pairs."""
    for key, child in iter_children_items(node):
        yield from _post_walk_items(child, __key__=key)
    yield __key__, node


def _post_walk_values(node: TreeLike) -> Iterable[Tuple[Any]]:
    """Create a post-order tree traversal iterator of values."""
    if (iter_children_values := getattr(node, "iter_children_values", None)) is not None:
        for child in iter_children_values():
            yield from _post_walk_values(child)
    yield node


post_walk_items = utils.as_xiter(_post_walk_items)
post_walk_values = utils.as_xiter(_post_walk_values)


def _bfs_walk_items(
    node: TreeLike, *, __key__: Optional[TreeKey] = None, __queue__: Optional[List] = None
) -> Iterable[Tuple[Optional[TreeKey], Any]]:
    """Create a tree traversal iterator of (key, value) pairs by tree levels (Breadth-First Search)."""
    __queue__ = __queue__ or []
    yield __key__, node
    __queue__.extend(iter_children_items(node))
    if __queue__:
        key, child = __queue__.pop(0)
        yield from _bfs_walk_items(child, __key__=key, __queue__=__queue__)


def _bfs_walk_values(
    node: TreeLike, *, __queue__: Optional[List] = None
) -> Iterable[Tuple[TreeKey, Any]]:
    """Create a tree traversal iterator of values by tree levels (Breadth-First Search)."""
    __queue__ = __queue__ or []
    yield node
    if (iter_children_values := getattr(node, "iter_children_values", None)) is not None:
        __queue__.extend(iter_children_values())
    if __queue__:
        child = __queue__.pop(0)
        yield from _bfs_walk_values(child, __queue__=__queue__)


bfs_walk_items = utils.as_xiter(_bfs_walk_items)
bfs_walk_values = utils.as_xiter(_bfs_walk_values)


def walk_items(
    node: TreeLike, traversal_order: TraversalOrder = TraversalOrder.PRE_ORDER
) -> utils.XIterable[Tuple[Optional[TreeKey], Any]]:
    """Create a tree traversal iterator of (key, value) pairs."""
    if traversal_order is traversal_order.PRE_ORDER:
        return pre_walk_items(node=node)
    elif traversal_order is traversal_order.POST_ORDER:
        return post_walk_items(node=node)
    elif traversal_order is traversal_order.BFS_ORDER:
        return bfs_walk_items(node=node)
    else:
        raise ValueError(f"Invalid '{traversal_order}' traversal order.")


def walk_values(
    node: TreeLike, traversal_order: TraversalOrder = TraversalOrder.PRE_ORDER
) -> utils.XIterable[Any]:
    """Create a tree traversal iterator of values."""
    if traversal_order is traversal_order.PRE_ORDER:
        return pre_walk_values(node=node)
    elif traversal_order is traversal_order.POST_ORDER:
        return post_walk_values(node=node)
    elif traversal_order is traversal_order.BFS_ORDER:
        return bfs_walk_values(node=node)
    else:
        raise ValueError(f"Invalid '{traversal_order}' traversal order.")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import collections
import copy
import operator
from typing import Generator, Optional, Type

import boltons.typeutils

import gt4py.cartesian.gtc.utils as gtc_utils
from gt4py import eve
from gt4py.cartesian import utils as gt_utils
from gt4py.cartesian.gtc import common

from .nodes import Location, Node


#: Marker value used to avoid confusion with `None`
#: (specially in contexts where `None` could be a valid value)
NOTHING = boltons.typeutils.make_sentinel(name="NOTHING", var_name="NOTHING")


def iter_attributes(node: Node):
    """
    Yield a tuple of ``(attrib_name, value)`` for each attribute in ``node.attributes``
    that is present on *node*.
    """
    for attrib_name in node.attributes:
        try:
            yield attrib_name, getattr(node, attrib_name)
        except AttributeError:
            pass


class IRNodeVisitor:
    def visit(self, node: Node, **kwargs):
        return self._visit(node, **kwargs)

    def _visit(self, node: Node, **kwargs):
        visitor = self.generic_visit
        if isinstance(node, Node):
            for node_class in node.__class__.__mro__:
                method_name = "visit_" + node_class.__name__
                if hasattr(self, method_name):
                    visitor = getattr(self, method_name)
                    break

        return visitor(node, **kwargs)

    def generic_visit(self, node: Node, **kwargs):
        items = []
        if isinstance(node, (str, bytes, bytearray)):
            pass
        elif isinstance(node, collections.abc.Mapping):
            items = node.items()
        elif isinstance(node, collections.abc.Iterable):
            items = enumerate(node)
        elif isinstance(node, Node):
            items = iter_attributes(node)
        else:
            pass

        for key, value in items:
            self._visit(value, **kwargs)


class IRNodeMapper:
    def visit(self, node: Node, **kwargs):
        return self._visit(node, **kwargs)

    def _visit(self, node: Node, **kwargs):
        visitor = self.generic_visit
        if isinstance(node, Node):
            for node_class in node.__class__.__mro__:
                method_name = "visit_" + node_class.__name__
                if hasattr(self, method_name):
                    visitor = getattr(self, method_name)
                    break

        return visitor(node, **kwargs)

    def generic_visit(self, node: Node, **kwargs):
        if isinstance(node, (str, bytes, bytearray)):
            return node
        elif isinstance(node, collections.abc.Iterable):
            if isinstance(node, collections.abc.Mapping):
                items = node.items()
            else:
                items = enumerate(node)
            setattr_op = operator.setitem
            delattr_op = operator.delitem
        elif isinstance(node, Node):
            items = iter_attributes(node)
            setattr_op = setattr
            delattr_op = delattr
        else:
            return node

        del_items = []
        for key, old_value in items:
            new_value = self._visit(old_value, **kwargs)
            if new_value == NOTHING:
                del_items.append(key)
            elif new_value != old_value:
                setattr_op(node, key, new_value)
        for key in reversed(del_items):  # reversed, so that keys remain valid in sequences
            delattr_op(node, key)

        return node


def iter_nodes_of_type(root_node: Node, node_type: Type) -> Generator[Node, None, None]:
    """Yield an iterator over the nodes of node_type inside root_node in DFS order."""

    def recurse(node: Node) -> Generator[Node, None, None]:
        for key, value in iter_attributes(node):
            if isinstance(node, collections.abc.Iterable):
                if isinstance(node, collections.abc.Mapping):
                    children = node.values()
                else:
                    children = node
            else:
                children = gtc_utils.listify(value)

            for value in children:
                if isinstance(value, Node):
                    yield from recurse(value)

            if isinstance(node, node_type):
                yield node

    yield from recurse(root_node)


def location_to_source_location(loc: Optional[Location]) -> Optional[eve.SourceLocation]:
    if loc is None or loc.line <= 0 or loc.column <= 0:
        return None
    return eve.SourceLocation(line=loc.line, column=loc.column, source=loc.scope)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gt4py.eve import NodeTranslator

from . import cuir


class FuseKernels(NodeTranslator):
    """Fuse consecutive kernels into single kernel for better launch overhead.

    Fuse kernels that are directly translated from OIR vertical loops into separate kernels if no synchronization is required.
    """

    def visit_Program(self, node: cuir.Program) -> cuir.Program:
        def is_parallel(kernel: cuir.Kernel) -> bool:
            parallel = [
                loop.loop_order == cuir.LoopOrder.PARALLEL for loop in kernel.vertical_loops
            ]
            assert all(parallel) or not any(parallel), "Mixed k-parallelism in kernel"
            return any(parallel)

        kernels = [self.visit(node.kernels[0])]
        previous_parallel = is_parallel(kernels[-1])
        previous_writes = (
            kernels[-1]
            .walk_values()
            .if_isinstance(cuir.AssignStmt)
            .getattr("left")
            .if_isinstance(cuir.FieldAccess)
            .getattr("name")
            .to_set()
        )
        for kernel in node.kernels[1:]:
            parallel = is_parallel(kernel)
            reads_with_offsets = (
                kernel.walk_values()
                .if_isinstance(cuir.FieldAccess)
                .filter(
                    lambda x, parallel=parallel: any(
                        off != 0 for off in x.offset.to_dict().values()
                    )
                    or (x.offset.to_dict()["k"] != 0 and parallel)
                )
                .getattr("name")
                .to_set()
            )
            new_writes = (
                kernel.walk_values()
                .if_isinstance(cuir.AssignStmt)
                .getattr("left")
                .if_isinstance(cuir.FieldAccess)
                .getattr("name")
                .to_set()
            )

            if previous_parallel != parallel or reads_with_offsets & previous_writes:
                kernels.append(self.visit(kernel))
                previous_writes = new_writes
            else:
                kernels[-1].vertical_loops += self.visit(kernel.vertical_loops)
                previous_writes |= new_writes
            previous_parallel = parallel

        return cuir.Program(
            name=node.name,
            params=node.params,
            positionals=node.positionals,
            temporaries=node.temporaries,
            kernels=kernels,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Dict

from gt4py import eve
from gt4py.cartesian.gtc import oir
from gt4py.cartesian.gtc.definitions import Extent
from gt4py.cartesian.gtc.passes.horizontal_masks import mask_overlap_with_extent
from gt4py.cartesian.gtc.passes.oir_optimizations.utils import compute_horizontal_block_extents


class NoFieldAccessPruning(eve.NodeTranslator):
    def visit_HorizontalExecution(self, node: oir.HorizontalExecution) -> Any:
        try:
            next(
                iter(
                    acc
                    for left in node.walk_values().if_isinstance(oir.AssignStmt).getattr("left")
                    for acc in left.walk_values().if_isinstance(oir.FieldAccess)
                )
            )
        except StopIteration:
            return eve.NOTHING

        return node

    def visit_VerticalLoopSection(self, node: oir.VerticalLoopSection) -> Any:
        horizontal_executions = self.visit(node.horizontal_executions)
        if not horizontal_executions:
            return eve.NOTHING
        return oir.VerticalLoopSection(
            interval=node.interval,
            horizontal_executions=horizontal_executions,
            loc=node.loc,
        )

    def visit_VerticalLoop(self, node: oir.VerticalLoop) -> Any:
        sections = self.visit(node.sections)
        if not sections:
            return eve.NOTHING
        return oir.VerticalLoop(
            loop_order=node.loop_order,
            sections=sections,
            caches=node.caches,
            loc=node.loc,
        )

    def visit_Stencil(self, node: oir.Stencil, **kwargs):
        vertical_loops = self.visit(node.vertical_loops, **kwargs)
        accessed_fields = (
            eve.walk_values(vertical_loops).if_isinstance(oir.FieldAccess).getattr("name").to_set()
        )
        declarations = [decl for decl in node.declarations if decl.name in accessed_fields]
        return oir.Stencil(
            name=node.name,
            vertical_loops=vertical_loops,
            params=node.params,
            declarations=declarations,
            loc=node.loc,
        )


class UnreachableStmtPruning(eve.NodeTranslator):
    def visit_Stencil(self, node: oir.Stencil) -> oir.Stencil:
        block_extents = compute_horizontal_block_extents(node)
        return self.generic_visit(node, block_extents=block_extents)

    def visit_HorizontalExecution(
        self, node: oir.HorizontalExecution, *, block_extents: Dict[int, Extent]
    ) -> oir.HorizontalExecution:
        return self.generic_visit(node, block_extent=block_extents[id(node)])

    def visit_HorizontalRestriction(
        self, node: oir.HorizontalRestriction, *, block_extent: Extent, **kwargs: Any
    ) -> Any:
        overlap = mask_overlap_with_extent(node.mask, block_extent)
        return eve.NOTHING if overlap is None else node
